/*
    Copyright (C) 2005, 2006  Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*******************************************************************************
    BUSINESS LOGIC LAYER
    --------------------
        Describes the drawing of the screens.

                                                            Sunday, 4 dec 2005.
*******************************************************************************/

#include "../logic/CHexEd.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

CScreen::CScreen() : uiLowLevel(){
}

CScreen::~CScreen(){

}


void CScreen::ShowError(unsigned int ErrorNumber, const char *fmt, ...){
     va_list args;
     char *buf = 0;
     unsigned long  maxx = 0,
                    _x = 0,
                    len = 0;
         //allocate memory
         maxx = GetMaxX();
         if(!needmemory(&buf, &_x, maxx + 1)) return;
         memset(buf, 0, maxx + 1);

         //header spaces
         if(ErrorNumber)
            sprintf(buf, "^ERROR %d: ", ErrorNumber);
         else
            sprintf(buf, "^ ");

         //output string
         va_start(args, fmt);
         vsnprintf(buf + strlen(buf), maxx - strlen(buf), fmt, args);
         va_end(args);

         //trailer spaces
         _x = strlen(buf);
         len = _x;
         if(len > maxx)len = maxx;
         while(_x <= maxx){ buf[_x] = 32; _x++; }

         //output to screen
         setDisplayModus(DISP_REVERSE);
         printxy(0, GetMaxY() - 1, buf);
     freememory(buf);
}

//clears line smoothly, safe formatted output
void CScreen::clrprintxy(int x, int y, const char *fmt, ...){
     va_list args;
     char *buf = 0;
     unsigned long  maxx = 0,
                    _x = 0,
                    len = 0;
         //allocate memory
         maxx = GetMaxX();
         if(!needmemory(&buf, &_x, maxx + 1)) return;
         memset(buf, 0, maxx + 1);

         //header spaces
         _x = 0;
         while(_x < (unsigned long)x && _x <= maxx){ buf[_x] = 32; _x++; }

         //output string
         va_start(args, fmt);
         vsnprintf(buf + strlen(buf), maxx - strlen(buf), fmt, args);
         va_end(args);

         //trailer spaces
         _x = strlen(buf);
         len = _x;
         if(len > maxx)len = maxx;
         while(_x < maxx){ buf[_x] = 32; _x++; }

         //output to screen
         printxy(x, y, buf);
         gotoxy(len, y);

     freememory(buf);
}


//safe formatted output right
void CScreen::right(int y, const char *fmt, ...){
     va_list args;
     char *buf = 0;
     unsigned long  maxx = 0,
                    _x = 0;
         //allocate memory
         maxx = GetMaxX();
         if(!needmemory(&buf, &_x, maxx + 1)) return;
         memset(buf, 0, maxx + 1);

         //output string
         va_start(args, fmt);
         vsnprintf(buf + strlen(buf), maxx - strlen(buf), fmt, args);
         va_end(args);

         //output to screen
         _x = strlen(buf);
         printxy(maxx - _x, y, buf);

     freememory(buf);
}

//safe formatted output center
void CScreen::center(int y, const char *fmt, ...){
     va_list args;
     char *buf = 0;
     unsigned long  maxx = 0,
                    _x = 0;
         //allocate memory
         maxx = GetMaxX();
         if(!needmemory(&buf, &_x, maxx + 1)) return;
         memset(buf, 0, maxx + 1);

         //output string
         va_start(args, fmt);
         vsnprintf(buf + strlen(buf), maxx - strlen(buf), fmt, args);
         va_end(args);

         //output to screen
         _x = strlen(buf);
         printxy((maxx - _x) / 2, y, buf);

     freememory(buf);
}

void CScreen::linexy(int x, int y, int len){
    gotoxy(x, y);
    while(len){ printchar(BOX_HORZ); len--; }
}

void CScreen::DrawUnknownScreen(SCREENID id){
    int cent = GetMaxY() >> 1;
    clrscr();
    setDisplayModus(DISP_REVERSE);
    center(cent - 2, "[CScreen.cpp::DrawScreen]");
    setDisplayModus(DISP_HIGHLIGHTED);
    center(cent - 1, "UNKNOWN SCREEN IDENTIFICATION");
    setDisplayModus(DISP_NORMAL);
    center(cent - 0, "Don't know how to draw this screen.");
    center(cent + 1, "SCREEN ID = %d", id);
}

void CScreen::DrawIntro(){
    int c = 0, y;
        //center y
        while(intro[c]) c++;
        y = (GetMaxY() - c) / 2;
        if(y < 0) y = 0;

        c = 0;
        while(intro[c]){
            if(y > GetMaxY()) break;
            if (c>5) setDisplayModus(DISP_HIGHLIGHTED); else setDisplayModus(DISP_NORMAL);
            if(c==0) setDisplayModus(DISP_REVERSE);
            if(c==1)
              #ifdef WIN32
                center(y, "Build: Windows");
              #endif
              #ifdef __UNIX__
                center(y, "Build: Unix/Linux");
              #endif
            else if(c==2) center(y, "Version %.1f", VERSION);
            else center(y, intro[c]);
            y++;
            c++;
        };

        setDisplayModus(DISP_NORMAL);
        center(0, "src/bin at www.geocities.com/hvansteenhuyse");
        clrprintxy(0, GetMaxY()-1, " hvansteenhuyse@yahoo.com");
        right(GetMaxY()-1, "C++ Code: sat 3, dec 2005 ");
}


void CScreen::DrawLS(ioDIR *dir){
    int sw_found = 0;
    char *spaces;
    long spc;
    char *SelectedFileName = "";
    CStr *FullPath;
    DIRFILE *df;
    unsigned long item;
    long    maxx = GetMaxX() - 2, maxy = GetMaxY() - 2, x, y,
            longest = 0, len;
    int     columns = 0, column;

    if(!dir || maxx < 0 || maxy < 0)return;

    if(strlen(dir->mSearchFor.get()) > 0 && dir->mSearchFor.get()[0] == ':'){
        y = 0;
        column = 0;
        spc = 0;
        setDisplayModus(DISP_REVERSE);
        clrprintxy(0, y, " LS COMMANDS: "); y++;
        setDisplayModus(DISP_NORMAL);
        clrprintxy(0, y, " :cd      change directory, e.g. :cd /mnt/share or :cd A:\\"); y++;
        y--;
    }else{
      if(strlen(dir->mSearchFor.get()) > 0){
            //zoek het element startend met mSearchFor
            item = dir->GetFirstFile();
            len = (long)strlen(dir->mSearchFor.get());
            while(!sw_found && item < dir->CountElements()){
                df = dir->ElementAt(item);
                if(strncmp(df->FileName, dir->mSearchFor.get(), len) == 0){
                    dir->SetSelectedItem(item);
                    sw_found = 1;
                }
                item++;
            }
      }
        //get longest filename
        item = 1;
        while(item < dir->CountElements()){
            df = dir->ElementAt(item);
            if(df){
                len = (long)strlen(df->FileName);
                if(longest < len) longest = len;
            }
            item++;
        }
        //aprroximate colums
        longest += 4; //4 spaces between each column
        if(longest > maxx) longest = maxx;
        columns = maxx / longest;
        if(columns < 1) columns = 1;

        if(dir->mLastKey == KHOME){
            dir->SubTopItem(dir->GetTopItem());
            dir->SubSelectedItem(dir->GetSelectedItem());
            dir->mSearchFor.clear();
        }

        //draw on screen
        for(int sw_draw = 0; sw_draw <= 1; sw_draw++){
            x = 1;
            y = 0;
            column = 1;
            if( dir->GetSelectedItem() > (unsigned long)(columns * (maxy - 1))
                &&
                //(dir->mLastKey == 9 || dir->mLastKey >= 32  && dir->mLastKey <= 126)
                strlen(dir->mSearchFor.get())
                &&
                (dir->GetTopItem() < dir->GetSelectedItem() - columns * (maxy - 1))
                ||
                (dir->GetTopItem() > dir->GetSelectedItem())
                )
                    dir->SetTopItem(dir->GetSelectedItem());

            item = dir->GetTopItem();
            printxy(0, y, " ");
            while(item < dir->CountElements()){
                if(!sw_draw){
                    if(column > columns){ x = 1; y++; column = 1; }
                    if(item == dir->GetSelectedItem()){
                        //process key input
                        if(dir->mLastKey == KPGD){
                            dir->mSearchFor.clear();
                            spc = maxy * columns;
                            if(dir->GetSelectedItem() + spc >= dir->CountElements()){
                                while(dir->GetSelectedItem() + spc >= dir->CountElements()) spc--;
                            }
                            dir->AddSelectedItem(spc);
                            if((long)((dir->GetSelectedItem() - dir->GetTopItem()) / columns) >= maxy - 1 )dir->AddTopItem(spc);
                        }
                        if(dir->mLastKey == KPGU){
                            dir->mSearchFor.clear();
                            dir->SubSelectedItem(maxy * columns);
                            dir->SubTopItem(maxy * columns);
                        }
                        if(dir->mLastKey == KRIGHT){
                            dir->mSearchFor.clear();
                            if(dir->GetSelectedItem() + 1 < dir->CountElements()){
                                dir->AddSelectedItem(1);
                                if(y == maxy -1 && column == columns) dir->AddTopItem(columns);
                            }
                        }
                        if(dir->mLastKey == KLEFT){
                            dir->mSearchFor.clear();
                            if(dir->GetSelectedItem() - 1 > 0){
                                dir->SubSelectedItem(1);
                                if(x == 1 && y == 0) dir->SubTopItem(columns);
                            }
                        }
                        if(dir->mLastKey == KDOWN){
                            dir->mSearchFor.clear();
                            dir->AddSelectedItem(columns);
                            if(y >= maxy - 1) dir->AddTopItem(columns);
                        }
                        if(dir->mLastKey == KUP){
                            dir->mSearchFor.clear();
                            dir->SubSelectedItem(columns);
                            while((long)(dir->GetTopItem() > dir->GetSelectedItem()) && dir->GetSelectedItem()) dir->SubTopItem(columns);
                        }
                        item = dir->CountElements();    //leave loop
                    }
                    if(y >= maxy)item = dir->CountElements();    //leave loop
                }else{
                    df = dir->ElementAt(item);
                    if(df){
                        setDisplayModus(DISP_NORMAL);
                        if(column > columns){ x = 1; y++; column = 1; printxy(0, y, " "); /*clrprintxy(0, y, ""); */}
                        if(df->IsFolder) setDisplayModus(DISP_HIGHLIGHTED);
                        if(item == dir->GetSelectedItem()){
                            setDisplayModus(DISP_REVERSE);
                            SelectedFileName = df->FileName;
                        }
                        if(y < maxy){
                            printxy(x, y, df->FileName);
                            if(column < columns)
                                spc = longest - strlen(df->FileName);
                            else
                                spc = maxx + 2 - x - strlen(df->FileName);
                            if(spc > 0){
                                spaces = (char*)malloc(spc + 2);
                                memset(spaces, ' ', spc);
                                spaces[spc] = 0;
                                setDisplayModus(DISP_NORMAL);
                                print(spaces);
                                free(spaces);
                            }
                        }else
                            item = dir->CountElements();    //leave loop
                        x += longest;
                    }
                }
                column++;
                item++;
            }
        }
    }
    setDisplayModus(DISP_NORMAL);

    spc = maxx + 3 - x;
    if(column > 1 && spc > 0){
        spaces = (char*)malloc(spc + 2);
        memset(spaces, ' ', spc);
        spaces[spc] = 0;
        print(spaces);
        free(spaces);
    }
    if(y < maxy){
      y++;
      while(y < maxy){
          clrprintxy(0, y, "~");
          y++;
      }
    }
    linexy(0, y, maxx + 2);
    y++;

    //display of selected file...
    if(strlen(dir->mSearchFor.get()) > 0 && dir->mSearchFor.get()[0] == ':'){
        setDisplayModus(DISP_HIGHLIGHTED);
        if((long)strlen(dir->mSearchFor.get()) > maxx + 1)
            spc = strlen(dir->mSearchFor.get()) - maxx - 1;
        else
            spc = 0;
        clrprintxy(0, y, dir->mSearchFor.get() + spc);
    }else{
        FullPath = new CStr();
        FullPath->append(dir->Directory);
        FullPath->append(ioLowLevel::ioGetFileSeparatorChar());
        FullPath->append((strlen(dir->mSearchFor.get())== 0 ? SelectedFileName : dir->mSearchFor.get()));
        if((long)strlen(FullPath->get()) > maxx + 1)
            spc = strlen(FullPath->get()) - maxx - 1;
        else
            spc = 0;
        clrprintxy(0, y, FullPath->get() + spc);
        delete FullPath;
    }
    dir->mLastKey = KNUL;
}

void CScreen::DrawSearch(CStr *search, int sw_hexside){
    unsigned long len = search->getsize(),
                  aScreen = GetHexDumpScreenSize(),
                  asciiX = (sw_hexside == 0 ? GetMinAsciiDump() : 0),
                  from = 0,
                  amount = 0;
    long    maxx = GetMaxX(), maxy = GetMaxY(), x = 0, y = 0, lines = 0;

    if(maxx < 0 || maxy < 0) return;

    if(len > aScreen) from = len - aScreen;

    if(search->getsize()){
        lines = HexDump(search->get() + from, search->get() + search->getsize(), 0);
        lines++; clrprintxy(0, lines, "~");
        lines++; clrprintxy(0, lines, "~");
    }else{
        clrprintxy(0, lines, "~"); lines++;
        clrprintxy(0, lines, "~"); lines++;
    }

    if(lines + 9 < maxy - 1){
        y = lines;
        while(y < maxy - 1){
            if(y == lines){
                setDisplayModus(DISP_REVERSE);
                clrprintxy(0, y, " SEARCH PATTERN:"); y++;
                setDisplayModus(DISP_NORMAL);
                clrprintxy(0, y, " <left> / <right>.......switch between hexadecimal or text input"); y++;
                clrprintxy(0, y, " <escape>...............back to dump"); y++;
                clrprintxy(0, y, " <backspace>............delete the tail-value"); y++;
                clrprintxy(0, y, " <enter>................find first occurence"); y++;
                clrprintxy(0, y, " After <enter> key was hit:"); y++;
                clrprintxy(0, y, "        <n> searches next occurence"); y++;
                clrprintxy(0, y, "        <p> searches previous occurence"); y++;
                setDisplayModus(DISP_REVERSE);
                clrprintxy(0, y, " Just start typing the search pattern...");
            }else{
                setDisplayModus(DISP_NORMAL);
                clrprintxy(0, y, "~");
            }
            y++;
        }
    }

    setDisplayModus(DISP_HIGHLIGHTED);
    clrprintxy(0, maxy - 1, "SEARCH: Expected input %s values, [#chrs: %d]", (sw_hexside ? "hexadecimal" : "text"), search->getsize());
    amount = len - from;
    y = amount / GetHexDumpLineSize();
    x = asciiX + amount % GetHexDumpLineSize();
    if(sw_hexside) x *= 3;
    if(x < 1) x = 1;
    if(y > maxy - 2){ y = maxy - 2;  x = GetHexDumpLineSize() * ((sw_hexside ? 1 : 0) * 3) ; }
    gotoxy(x, y);
}

void CScreen::DrawCommandLine(char *cmd, int sw_reverse){
    int y = GetMaxY() - 1,
        x = GetMaxX() - 1;
    unsigned long len = strlen(cmd);
    if(y > 0){
      setDisplayModus((sw_reverse ? DISP_REVERSE : DISP_HIGHLIGHTED));
      if(len){
          if(len > (unsigned long)x) cmd += len - x;
          clrprintxy(0, y, cmd);
      }
      else
          clrprintxy(0, y, cmd);
    }
}

//returns maximum X coordination of the hexdump part on the screen
long CScreen::GetMaxHexDump(){
    return 1 + GetHexDumpLineSize() * 3;
}

//returns minimum X coordination of the ascii dump part on the screen
long CScreen::GetMinAsciiDump(){
    return GetMaxX() - 1 - GetHexDumpLineSize();
}

/*computes the amount of chars required to fill a hexdump line on the screen*/
long CScreen::GetHexDumpLineSize(){
    long maxx, chars = 0;
      maxx = GetMaxX() - 2;
      if(maxx > 0) chars = maxx >> 2;   //(#chars on line - 2) divide by 4.
                                        // -2:= 1 blank on left and right of the screen
                                        // 4 := 2 digits hex + 1 space + 1 ascii
      //log("maxx %d, chars = %d", maxx >> 2, chars);
    return (chars < 0 ? 0 : chars);
}

/*computes the amount of chars required to fill a whole hexdump screen*/
long CScreen::GetHexDumpScreenSize(){
    long maxy, chars = 0;
      maxy = GetMaxY() - 1;
      if(maxy > 0) chars = GetHexDumpLineSize() * maxy;
      //log("maxy %d, chars = %d", maxy, chars);
    return chars;
}


int CScreen::HexDump(char *from, char *to, int swClear){
    int lines = 0;
    char *scrLine = 0,
         *chrs = 0,
         ch;
    long y = 0,
        ctel,
        maxx,
        maxy,
        chars;

    setDisplayModus(DISP_NORMAL);
    maxx = GetMaxX();
    maxy = GetMaxY() - 1;

    chars =  GetHexDumpLineSize();
    if(chars > 0){
        scrLine = (char*)malloc(maxx + 1);
        chrs = (char*)malloc(maxx + 1);
        if(scrLine == 0 || chrs == 0){
          clrscr();
          ShowError(errno, "[CScreen::HexDump] Memory Allocation Failed.");
        }else{
            while(y < maxy){
                memset(scrLine, 0, maxx + 1);
                memset(chrs, 0, maxx + 1);
                gotoxy(0, y);
                sprintf(scrLine, " ");
                if(from < to){
                    /*de dump van de karakters*/
                    ctel = 0;
                    while(ctel < chars){
                        if(from < to){
                            ch = from[0];
                            atoh(ch, scrLine + strlen(scrLine));
                            sprintf(scrLine + strlen(scrLine), " ");
                            ch = GetDisplayChar(ch);
                            sprintf(chrs + strlen(chrs), "%c", ch);
                            from++;
                        }else{
                            sprintf(scrLine + strlen(scrLine), "   ");
                            sprintf(chrs + strlen(chrs), " ");
                        }
                        ctel++;
                        lines = y;
                    }
                    while((int)strlen(scrLine) < maxx - chars - 1) sprintf(scrLine + strlen(scrLine), " ");
                    sprintf(scrLine + strlen(scrLine), "%s ", chrs);
                    print(scrLine);
                }else{
                    if(swClear)clrprintxy(0, y, "~");
                }
                y++;
            }
        }
    }
    if(scrLine)free(scrLine);
    if(chrs)free(chrs);
    return lines;
}

void CScreen::DrawHelp(){
    int sw = 0, c = 0, maxy = GetMaxY() - 1;
        if(maxy < 1) return;
        while(c < maxy){
            if(!sw && help[c]){
                if(c == 0)setDisplayModus(DISP_REVERSE); else setDisplayModus(DISP_NORMAL);
                clrprintxy(0, c, help[c]);
            }else{
                sw = 1;
                clrprintxy(0, c, "~");
            }
            c++;
        }
    setDisplayModus(DISP_HIGHLIGHTED);
    clrprintxy(0, maxy, " press any key...");

}

